from typing import Protocol
from collections.abc import Callable, Iterable
from dataclasses import dataclass

import xarray as xr


class Regionalizer(Protocol):
    """
    Use to regionalize gridded, n-dimentional dataset.

    The key is that it can be loaded with invariants or state information (e.g. weights, spatial geometry) before it is used in the transformation process to regionalize gridded data.
    """

    def regionalize(self, ds: xr.Dataset) -> xr.Dataset: ...


# TODO: Can we still read docstr of callables through their attributes after theyve been passed to an instantiated TransformationStrategy?
@dataclass(frozen=True)
class TransformationStrategy:
    """
    Named, tranformation steps applied to input gridded data, pre/post regionalization, to create a derived variable as output.

    These steps should be general. They may contain logic for sanity checks on inputs and outputs, calculating derived variables and climate indices, adding or checking metadata or units. Avoid including logic for cleaning, or harmonizing input data, especially if it is specific to a single project's usecase. Generally avoid using a single strategy to output multiple unrelated variables.
    """

    preprocess: Callable[[xr.Dataset], xr.Dataset]
    postprocess: Callable[[xr.Dataset], xr.Dataset]


# Use class for segment weights because we're making assumptions/enforcements about the weight data's content and interactions...
class SegmentWeights:
    """
    Segment weights to regionalize regularly-gridded data
    """

    def __init__(self, weights: xr.Dataset):
        target_variables = ("lat", "lon", "weight", "region")
        missing_variables = [v for v in target_variables if v not in weights.variables]
        if missing_variables:
            raise ValueError(
                f"input weights is missing required {missing_variables} variable(s)"
            )
        self._data = weights

    def regionalize(self, x: xr.Dataset) -> xr.Dataset:
        """
        Regionalize input gridded data
        """
        # TODO: See how this errors in different common scenarios. What happens on the
        #  unhappy path?
        region_sel = x.sel(lat=self._data["lat"], lon=self._data["lon"])
        out = (region_sel * self._data["weight"]).groupby(self._data["region"]).sum()
        # TODO: Maybe drop lat/lon and set 'region' as dim/coord? I feel like we can do
        #  this because we're asking weights to strictly match input's lat/lon. Maybe
        #  make this a req of segment weights we're reading in?
        return out

    def __call__(self, x: xr.Dataset) -> xr.Dataset:
        return self.regionalize(x)


def _default_transform_merge(x: Iterable[xr.Dataset]) -> xr.Dataset:
    return xr.merge(x)


def apply_transformations(
    gridded: xr.Dataset,
    *,
    strategies: Iterable[TransformationStrategy],
    regionalize: Callable[[xr.Dataset], xr.Dataset],
    merge_transformed: Callable[[Iterable[xr.Dataset]], xr.Dataset] | None = None,
) -> xr.Dataset:
    """
    Apply multiple regionalized transformations output to a single Dataset.
    """
    strategies = tuple(strategies)

    if merge_transformed is None:
        merge_transformed = _default_transform_merge

    transformed = []
    for s in strategies:
        preprocessed = s.preprocess(gridded)
        regionalized = regionalize(preprocessed)
        postprocessed = s.postprocess(regionalized)
        transformed.append(postprocessed)

    return merge_transformed(transformed)



from typing import Any
from collections.abc import Callable, Iterable
from dataclasses import dataclass

import xarray as xr


def _default_predictors_parameters_merge(x: Iterable[xr.Dataset]) -> xr.Dataset:
    return xr.merge(x)


@dataclass(frozen=True)
class Projector:
    """
    Model to project effects, impacts and/or damages.
    """

    preprocess: Callable[[xr.Dataset], Any]
    project: Callable[[Any], Any]
    postprocess: Callable[[Any], xr.Dataset]


def project(
    predictors: xr.Dataset,
    *,
    model: Projector,
    parameters: xr.Dataset,
    merge_predictors_parameters: Callable[[Iterable[xr.Dataset]], xr.Dataset]
    | None = None,
) -> xr.Dataset:
    """
    Project given predictors, a model, and model parameters.
    """
    if merge_predictors_parameters is None:
        merge_predictors_parameters = _default_predictors_parameters_merge

    # Include model artifacts/params/coefs on input xr.Dataset so broadcasting/index problems are hit early... Also makes testing models easier.
    projection_input = merge_predictors_parameters([predictors, parameters])

    preprocessed = model.preprocess(projection_input)
    projected = model.project(preprocessed)
    postprocessed = model.postprocess(projected)

    return postprocessed


# Think about this vs pickling (or use joblib) a model with parameters into an artifact that gets passed around.

# We can use this same project() + Projector pattern to model impacts and then damages, if needed.



from muuttaa.core.project import Projector, project  # noqa: F401
from muuttaa.core.transform import (
    TransformationStrategy,  # noqa: F401
    SegmentWeights,  # noqa: F401
    apply_transformations,  # noqa: F401
)



import numpy as np
import xarray as xr

from muuttaa import TransformationStrategy, SegmentWeights, apply_transformations


def test_apply_transformationstrategy():
    """
    Create simple transformationstrategy and regionalization function, test basic application.
    """
    input_ds = xr.Dataset({"variable1": (["idx"], [0, 0, 0])})
    expected = xr.Dataset({"variable1": (["idx"], [13.5, 13.5, 13.5])})

    # Each of the transformation steps should add to the variable.
    # We'll know something basic is off if it doesn't add to `expected`.
    def _pre(x):
        return x[["variable1"]] + 1

    def _post(x):
        return x[["variable1"]] + 10

    test_transform = TransformationStrategy(preprocess=_pre, postprocess=_post)

    def fake_regionalization(x):
        return x[["variable1"]] + 2.5

    output = apply_transformations(
        input_ds,
        strategies=[test_transform],
        regionalize=fake_regionalization,
    )

    xr.testing.assert_allclose(output, expected)


def test_segmentweight_regionalization():
    """
    Basic test calling SegmentWeight for regionalization
    """
    # Define some basics so we can instantiate weights and data
    # to regionalize.
    da = xr.DataArray(
        np.arange(25).reshape([5, 5]),
        dims=("lon", "lat"),
        coords={
            "lon": np.arange(5),
            "lat": np.arange(5),
        },
        name="variable1",
    )
    w = xr.Dataset(
        {
            "region": (["idx"], ["a", "a", "a", "b"]),
            "weight": (["idx"], [0.3, 0.3, 0.3, 1.0]),
            "lon": (["idx"], [2, 3, 4, 1]),
            "lat": (["idx"], [0, 0, 0, 2]),
        },
    )

    expected = xr.DataArray(
        np.array([13.5, 7.0]),
        dims="region",
        coords={
            "region": ["a", "b"],
        },
        name="variable1",
    ).to_dataset()

    weights = SegmentWeights(w)
    actual = weights(da.to_dataset())

    xr.testing.assert_allclose(actual, expected)


def test_segmentweight_regionalization_extradim():
    """
    Basic SegmentWeight regionalization test, but if input data has extra time dim.
    """
    # Define inputs so we can instantiate weights and have data to regionalize.
    # Giving da extra time dim...
    da = xr.DataArray(
        np.arange(125).reshape([5, 5, 5]),
        dims=("lon", "lat", "time"),
        coords={
            "lon": np.arange(5),
            "lat": np.arange(5),
            "time": np.arange(5),
        },
        name="variable1",
    )
    w = xr.Dataset(
        {
            "region": (["idx"], ["a", "a", "a", "b"]),
            "weight": (["idx"], [0.3, 0.6, 0.3, 1.0]),
            "lon": (["idx"], [2, 3, 4, 1]),
            "lat": (["idx"], [0, 0, 0, 2]),
        },
    )

    expected = xr.DataArray(
        np.array([[90.0, 91.2, 92.4, 93.6, 94.8], [35.0, 36.0, 37.0, 38.0, 39.0]]),
        dims=("region", "time"),
        coords={
            "region": ["a", "b"],
            "time": np.arange(5),
        },
        name="variable1",
    ).to_dataset()

    weights = SegmentWeights(w)
    actual = weights(da.to_dataset())

    xr.testing.assert_allclose(actual, expected)



import xarray as xr

from muuttaa import Projector, project


def test_basic_projection():
    """
    Basic test running a Projector with project().
    """
    predictors = xr.Dataset({"foobar": (["idx"], [0, 0, 0])})
    params = xr.Dataset({"ni": (["idx"], [1, 2, 3])})
    expected = xr.Dataset({"impact": (["idx"], [13, 14, 15])})

    def _pre(x):
        out = xr.Dataset()
        out["foobar"] = x["foobar"] + 1
        out["ni"] = x["ni"]
        return out

    def _post(x):
        return x[["impact"]] + 10

    def _model(x):
        return (x["foobar"] * 2 + x["ni"]).to_dataset(name="impact")

    test_impact_model = Projector(
        preprocess=_pre,
        project=_model,
        postprocess=_post,
    )

    actual = project(predictors, model=test_impact_model, parameters=params)
    xr.testing.assert_allclose(actual, expected)



from muuttaa._version import __version__  # noqa: F401
from muuttaa.core import (
    TransformationStrategy,  # noqa: F401
    SegmentWeights,  # noqa: F401
    apply_transformations,  # noqa: F401
    Projector,  # noqa: F401
    project,  # noqa: F401
)

__all__ = [
    "TransformationStrategy",
    "SegmentWeights",
    "apply_transformations",
    "Projector",
    "project",
]
